'''
Triplet loss modified from https://github.com/omoindrot/tensorflow-triplet-loss/blob/master/model/triplet_loss.py
Author: Olivier Moindrot from Stanford
'''

import tensorflow as tf


def _pairwise_distances(embeddings, squared=False):
    """Compute the 2D matrix of distances between all the embeddings.
    Args:
        embeddings: tensor of shape (batch_size, embed_dim)
        squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                 If false, output is the pairwise euclidean distance matrix.
    Returns:
        pairwise_distances: tensor of shape (batch_size, batch_size)
    """
    # Get the dot product between all embeddings
    # shape (batch_size, batch_size)
    dot_product = tf.matmul(embeddings, tf.transpose(embeddings))

    # Get squared L2 norm for each embedding. We can just take the diagonal of `dot_product`.
    # This also provides more numerical stability (the diagonal of the result will be exactly 0).
    # shape (batch_size,)
    square_norm = tf.diag_part(dot_product)

    # Compute the pairwise distance matrix as we have:
    # ||a - b||^2 = ||a||^2  - 2 <a, b> + ||b||^2
    # shape (batch_size, batch_size)
    distances = tf.expand_dims(square_norm, 1) - 2.0 * dot_product + tf.expand_dims(square_norm, 0)

    # Because of computation errors, some distances might be negative so we put everything >= 0.0
    distances = tf.maximum(distances, 0.0)

    if not squared:
        # Because the gradient of sqrt is infinite when distances == 0.0 (ex: on the diagonal)
        # we need to add a small epsilon where distances == 0.0
        mask = tf.to_float(tf.equal(distances, 0.0))
        distances = distances + mask * 1e-16

        distances = tf.sqrt(distances)

        # Correct the epsilon added: set the distances on the mask to be exactly 0.0
        distances = distances * (1.0 - mask)

    return distances


def _get_anchor_positive_triplet_mask(labels):
    """Return a 2D mask where mask[a, p] is True iff a and p are distinct and have same label.
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    Returns:
        mask: tf.bool `Tensor` with shape [batch_size, batch_size]
    """
    # Check that i and j are distinct
    indices_equal = tf.cast(tf.eye(tf.shape(labels)[0]), tf.bool)
    indices_not_equal = tf.logical_not(indices_equal)

    # Check if labels[i] == labels[j]
    # Uses broadcasting where the 1st argument has shape (1, batch_size) and the 2nd (batch_size, 1)
    labels_equal = tf.equal(tf.expand_dims(labels, 0), tf.expand_dims(labels, 1))

    # Combine the two masks
    mask = tf.logical_and(indices_not_equal, labels_equal)

    return mask


def _get_anchor_negative_triplet_mask(labels):
    """Return a 2D mask where mask[a, n] is True iff a and n have distinct labels.
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    Returns:
        mask: tf.bool `Tensor` with shape [batch_size, batch_size]
    """
    # Check if labels[i] != labels[k]
    # Uses broadcasting where the 1st argument has shape (1, batch_size) and the 2nd (batch_size, 1)
    labels_equal = tf.equal(tf.expand_dims(labels, 0), tf.expand_dims(labels, 1))

    mask = tf.logical_not(labels_equal)

    return mask


def _get_triplet_mask(labels):
    """Return a 3D mask where mask[a, p, n] is True iff the triplet (a, p, n) is valid.
    A triplet (i, j, k) is valid if:
        - i, j, k are distinct
        - labels[i] == labels[j] and labels[i] != labels[k]
    Args:
        labels: tf.int32 `Tensor` with shape [batch_size]
    """
    # Check that i, j and k are distinct
    indices_equal = tf.cast(tf.eye(tf.shape(labels)[0]), tf.bool)
    indices_not_equal = tf.logical_not(indices_equal)
    i_not_equal_j = tf.expand_dims(indices_not_equal, 2)
    i_not_equal_k = tf.expand_dims(indices_not_equal, 1)
    j_not_equal_k = tf.expand_dims(indices_not_equal, 0)

    distinct_indices = tf.logical_and(tf.logical_and(i_not_equal_j, i_not_equal_k), j_not_equal_k)

    # Check if labels[i] == labels[j] and labels[i] != labels[k]
    label_equal = tf.equal(tf.expand_dims(labels, 0), tf.expand_dims(labels, 1))
    i_equal_j = tf.expand_dims(label_equal, 2)
    i_equal_k = tf.expand_dims(label_equal, 1)

    valid_labels = tf.logical_and(i_equal_j, tf.logical_not(i_equal_k))

    # Combine the two masks
    mask = tf.logical_and(distinct_indices, valid_labels)

    return mask


def batch_all_triplet_loss(labels, embeddings, margin, squared=False):
    """Build the triplet loss over a batch of embeddings.
    We generate all the valid triplets and average the loss over the positive ones.
    Args:
        labels: labels of the batch, of size (batch_size,)
        embeddings: tensor of shape (batch_size, embed_dim)
        margin: margin for triplet loss
        squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                 If false, output is the pairwise euclidean distance matrix.
    Returns:
        triplet_loss: scalar tensor containing the triplet loss
    """
    # Get the pairwise distance matrix
    pairwise_dist = _pairwise_distances(embeddings, squared=squared)

    # shape (batch_size, batch_size, 1)
    anchor_positive_dist = tf.expand_dims(pairwise_dist, 2)
    assert anchor_positive_dist.shape[2] == 1, "{}".format(anchor_positive_dist.shape)
    # shape (batch_size, 1, batch_size)
    anchor_negative_dist = tf.expand_dims(pairwise_dist, 1)
    assert anchor_negative_dist.shape[1] == 1, "{}".format(anchor_negative_dist.shape)

    # Compute a 3D tensor of size (batch_size, batch_size, batch_size)
    # triplet_loss[i, j, k] will contain the triplet loss of anchor=i, positive=j, negative=k
    # Uses broadcasting where the 1st argument has shape (batch_size, batch_size, 1)
    # and the 2nd (batch_size, 1, batch_size)
    triplet_loss = anchor_positive_dist - anchor_negative_dist + margin

    # Put to zero the invalid triplets
    # (where label(a) != label(p) or label(n) == label(a) or a == p)
    mask = _get_triplet_mask(labels)
    mask = tf.to_float(mask)
    triplet_loss = tf.multiply(mask, triplet_loss)

    # Remove negative losses (i.e. the easy triplets)
    triplet_loss = tf.maximum(triplet_loss, 0.0)

    # Count number of positive triplets (where triplet_loss > 0)
    valid_triplets = tf.to_float(tf.greater(triplet_loss, 1e-16))
    num_positive_triplets = tf.reduce_sum(valid_triplets)
    num_valid_triplets = tf.reduce_sum(mask)
    fraction_positive_triplets = num_positive_triplets / (num_valid_triplets + 1e-16)

    # Get final mean triplet loss over the positive valid triplets
    triplet_loss = tf.reduce_sum(triplet_loss) / (num_positive_triplets + 1e-16)

    return triplet_loss, fraction_positive_triplets


def batch_hard_triplet_loss(labels, embeddings, margin, squared=False):
    """Build the triplet loss over a batch of embeddings.
    For each anchor, we get the hardest positive and hardest negative to form a triplet.
    Args:
        labels: labels of the batch, of size (batch_size,)
        embeddings: tensor of shape (batch_size, embed_dim)
        margin: margin for triplet loss
        squared: Boolean. If true, output is the pairwise squared euclidean distance matrix.
                 If false, output is the pairwise euclidean distance matrix.
    Returns:
        triplet_loss: scalar tensor containing the triplet loss
    """
    # Get the pairwise distance matrix
    pairwise_dist = _pairwise_distances(embeddings, squared=squared)

    # For each anchor, get the hardest positive
    # First, we need to get a mask for every valid positive (they should have same label)
    mask_anchor_positive = _get_anchor_positive_triplet_mask(labels)
    mask_anchor_positive = tf.to_float(mask_anchor_positive)

    # We put to 0 any element where (a, p) is not valid (valid if a != p and label(a) == label(p))
    anchor_positive_dist = tf.multiply(mask_anchor_positive, pairwise_dist)

    # shape (batch_size, 1)
    hardest_positive_dist = tf.reduce_max(anchor_positive_dist, axis=1, keepdims=True)
    tf.summary.scalar("hardest_positive_dist", tf.reduce_mean(hardest_positive_dist))

    # For each anchor, get the hardest negative
    # First, we need to get a mask for every valid negative (they should have different labels)
    mask_anchor_negative = _get_anchor_negative_triplet_mask(labels)
    mask_anchor_negative = tf.to_float(mask_anchor_negative)

    # We add the maximum value in each row to the invalid negatives (label(a) == label(n))
    max_anchor_negative_dist = tf.reduce_max(pairwise_dist, axis=1, keepdims=True)
    anchor_negative_dist = pairwise_dist + max_anchor_negative_dist * (1.0 - mask_anchor_negative)

    # shape (batch_size,)
    hardest_negative_dist = tf.reduce_min(anchor_negative_dist, axis=1, keepdims=True)
    tf.summary.scalar("hardest_negative_dist", tf.reduce_mean(hardest_negative_dist))

    # Combine biggest d(a, p) and smallest d(a, n) into final triplet loss
    triplet_loss = tf.maximum(hardest_positive_dist - hardest_negative_dist + margin, 0.0)

    # Get final mean triplet loss
    triplet_loss = tf.reduce_mean(triplet_loss)

    return triplet_loss



import tensorflow as tf
from .triplet_loss import batch_hard_triplet_loss

""" model structure of MUSE """


def structured_embedding(x, y, label_x, label_y, dim_z, triplet_margin,
                         n_hidden, keep_prob, weight_penalty, triplet_lambda):
    """
    Construct structure and loss function of MUSE

    Parameters:
        x:              input batches for transcript modality; matrix of  n * p, where n = batch size, p = number of genes
        y:              input batches for morphological modality; matrix of n * q, where n = batch size, q is the feature dimension
        label_x:        input sample labels for modality x
        label_y:        input sample labels for modality y
        dim_z:          dimension of joint latent representation
        triplet_margin: margin for triplet loss
        n_hidden:       hidden node number for encoder and decoder layers
        keep_prob:      keep probability for dropout in learning
        weight_penalty: weight for sparse penalty
        triplet_lambda: weight for triplet loss

    Outputs:
        z:              joint latent representations (n * dim_z)
        x_hat:          reconstructed x (same shape as x)
        y_hat:          reconstructed y (same shape as y)
        encode_x:       latent representation for modality x
        encode_y:       latent representation for modality y
        loss:           total loss
        reconstruct_loss: reconstruction loss
        sparse_penalty: sparse penalty
        trip_loss_x:    triplet loss for x
        trip_loss_y:    triplet loss for y

    Altschuler & Wu Lab 2020.
    Software provided as is under MIT License.
    """

    # encoder
    z, encode_x, encode_y = multiview_encoder(x, y, n_hidden, dim_z)

    # decoder
    w_init = tf.initializers.random_normal()
    w_x = tf.get_variable('w_selection_x', [z.get_shape()[1], z.get_shape()[1]], initializer=w_init)
    w_y = tf.get_variable('w_selection_y', [z.get_shape()[1], z.get_shape()[1]], initializer=w_init)

    with tf.variable_scope("decoder_x"):
        z_x = tf.matmul(z, w_x)
        x_hat = decoder(z_x, n_hidden, x.get_shape()[1], keep_prob)

    with tf.variable_scope("decoder_y"):
        z_y = tf.matmul(z, w_y)
        y_hat = decoder(z_y, n_hidden, y.get_shape()[1], keep_prob)

    # sparse penalty
    sparse_x = tf.sqrt(tf.reduce_sum(tf.reduce_sum(tf.square(w_x), axis=1)))
    sparse_y = tf.sqrt(tf.reduce_sum(tf.reduce_sum(tf.square(w_y), axis=1)))
    sparse_penalty = sparse_x + sparse_y

    # reconstruction errors (for x modality, only non-zero entries were used)
    x_mask = tf.sign(x)
    reconstruct_x = tf.reduce_sum(tf.norm(tf.multiply(x_mask, (x_hat - x)))) / tf.reduce_sum(x_mask)
    reconstruct_y = tf.reduce_mean(tf.norm(y_hat - y))
    reconstruct_loss = reconstruct_x + reconstruct_y

    # triplet errors
    trip_loss_x = batch_hard_triplet_loss(label_x, z, triplet_margin)
    trip_loss_y = batch_hard_triplet_loss(label_y, z, triplet_margin)

    loss = reconstruct_loss + weight_penalty * sparse_penalty \
           + triplet_lambda * trip_loss_x \
           + triplet_lambda * trip_loss_y

    return z, x_hat, y_hat, encode_x, encode_y, \
           loss, reconstruct_loss, sparse_penalty, trip_loss_x, trip_loss_y


def multiview_encoder(x, y, n_hidden, dim_z):
    """
    Encoder combines x and y to a joint latent representation

    Parameters:
        x:              input batches for transcript modality; matrix of  n * p, where n = batch size, p = number of genes
        y:              input batches for morphological modality; matrix of n * q, where n = batch size, q is the feature dimension
        n_hidden:       hidden node number for encoder and decoder layers
        dim_z:          dimension of joint latent representations

    Outputs:
        latent:         joint latent representations
        h_x:            latent representation for modality x
        h_y:            latent representation for modality y

    Altschuler & Wu Lab 2020.
    Software provided as is under MIT License.
    """

    # encoder for each modality
    with tf.variable_scope("encoder_x"):
        h_x = encoder(x, n_hidden)
    with tf.variable_scope("encoder_y"):
        h_y = encoder(y, n_hidden)

    # combine h_x and h_y to joint latent representation
    w_init = tf.contrib.layers.variance_scaling_initializer()
    b_init = tf.constant_initializer(0.)
    h = tf.concat([h_x, h_y], 1)
    wo = tf.get_variable('wo', [h.get_shape()[1], dim_z], initializer=w_init)
    bo = tf.get_variable('bo', [dim_z], initializer=b_init)
    latent = tf.matmul(h, wo) + bo

    return latent, h_x, h_y


def encoder(x, n_hidden):
    """
    Encoder for single modality

    Parameters:
        x:              input batches of single modality
        n_hidden:       hidden node number

    Outputs:
        o:              latent representation for single modality

    Altschuler & Wu Lab 2020.
    Software provided as is under MIT License.
    """

    # initializers
    w_init = tf.contrib.layers.variance_scaling_initializer()
    b_init = tf.constant_initializer(0.)

    # 1st hidden layer
    w0 = tf.get_variable('w0_e', [x.get_shape()[1], n_hidden], initializer=w_init)
    b0 = tf.get_variable('b0_e', [n_hidden], initializer=b_init)
    h0 = tf.matmul(x, w0) + b0
    h0 = tf.nn.elu(h0)

    # 2nd hidden layer
    w1 = tf.get_variable('w1_e', [h0.get_shape()[1], n_hidden], initializer=w_init)
    b1 = tf.get_variable('b1_e', [n_hidden], initializer=b_init)
    h1 = tf.matmul(h0, w1) + b1
    o = tf.nn.tanh(h1)

    return o


def decoder(z, n_hidden, n_output, keep_prob):
    """
    Decoder for single modality

    Parameters:
        z:              latent representation for single modality
        n_hidden:       hidden node number in decoder
        n_output:       feature dimension of original data
        keep_prob:      keep probability

    Outputs:
        y:              reconstructed feature

    Altschuler & Wu Lab 2020.
    Software provided as is under MIT License.
    """

    # initializers
    w_init = tf.contrib.layers.variance_scaling_initializer()
    b_init = tf.constant_initializer(0.)

    # 1st hidden layer
    w0 = tf.get_variable('w0_d', [z.get_shape()[1], n_hidden], initializer=w_init)
    b0 = tf.get_variable('b0_d', [n_hidden], initializer=b_init)
    h0 = tf.matmul(z, w0) + b0
    h0 = tf.nn.elu(h0)
    h0 = tf.nn.dropout(h0, keep_prob)

    # 2nd hidden layer
    w1 = tf.get_variable('w1_d', [h0.get_shape()[1], n_hidden], initializer=w_init)
    b1 = tf.get_variable('b1_d', [n_hidden], initializer=b_init)
    h1 = tf.matmul(h0, w1) + b1
    h1 = tf.nn.tanh(h1)
    h1 = tf.nn.dropout(h1, keep_prob)

    # output layer-mean
    wo = tf.get_variable('wo_d', [h1.get_shape()[1], n_output], initializer=w_init)
    bo = tf.get_variable('bo_d', [n_output], initializer=b_init)
    y = tf.matmul(h1, wo) + bo

    return y



import numpy as np
import tensorflow as tf
import .muse_architecture as muse_model
from scipy.spatial.distance import pdist
import phenograph

""" Model fitting and feature prediction of MUSE """


def muse_fit_predict(data_x,
                     data_y,
                     label_x,
                     label_y,
                     latent_dim=100,
                     n_epochs=500,
                     lambda_regul=5,
                     lambda_super=5):
    """
        MUSE model fitting and predicting:
    	  This function is used to train the MUSE model on multi-modality data

        Parameters:
          data_x:       input for transcript modality; matrix of  n * p, where n = number of cells, p = number of genes.
          data_y:       input for morphological modality; matrix of n * q, where n = number of cells, q is the feature dimension.
          label_x:      initial reference cluster label for transcriptional modality.
          label_y:      inital reference cluster label for morphological modality.
          latent_dim:   feature dimension of joint latent representation.
          n_epochs:     maximal epoch used in training.
          lambda_regul: weight for regularization term in the loss function.
          lambda_super: weight for supervised learning loss in the loss function.

        Output:
          latent:       joint latent representation learned by MUSE.
          reconstruct_x:reconstructed feature matrix corresponding to input data_x.
          reconstruct_y:reconstructed feature matrix corresponding to input data_y.
          latent_x:     modality-specific latent representation corresponding to data_x.
          latent_y:     modality-specific latent representation corresponding to data_y.

        Altschuler & Wu Lab 2020.
        Software provided as is under MIT License.
    """

    """ initial parameter setting """
    # parameter setting for neural network
    n_hidden = 128  # number of hidden node in neural network
    learn_rate = 1e-4  # learning rate in the optimization
    batch_size = 64  # number of cells in the training batch
    n_epochs_init = 200  # number of training epoch in model initialization
    print_epochs = 50  # epoch interval to display the current training loss
    cluster_update_epoch = 200  # epoch interval to update modality-specific clusters
    keep_prob_val = 1  # keep probability in training dropout

    # read data-specific parameters from inputs
    feature_dim_x = data_x.shape[1]
    feature_dim_y = data_y.shape[1]
    n_sample = data_x.shape[0]

    # GPU configuration
    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True

    """ construct computation graph using TensorFlow """
    tf.reset_default_graph()

    # raw data from two modalities
    x = tf.placeholder(tf.float32, shape=[None, feature_dim_x], name='input_x')
    y = tf.placeholder(tf.float32, shape=[None, feature_dim_y], name='input_y')

    # labels inputted for references
    ref_label_x = tf.placeholder(tf.float32, shape=[None], name='ref_label_x')
    ref_label_y = tf.placeholder(tf.float32, shape=[None], name='ref_label_y')

    # hyperparameter in triplet loss
    triplet_lambda = tf.placeholder(tf.float32, name='triplet_lambda')
    triplet_margin = tf.placeholder(tf.float32, name='triplet_margin')

    # dropout
    keep_prob = tf.placeholder(tf.float32, name='keep_prob')

    # network architecture
    z, x_hat, y_hat, encode_x, encode_y, loss, \
    reconstruction_error, weight_penalty, \
    trip_loss_x, trip_loss_y = muse_model.structured_embedding(x,
                                                               y,
                                                               ref_label_x,
                                                               ref_label_y,
                                                               latent_dim,
                                                               triplet_margin,
                                                               n_hidden,
                                                               keep_prob=keep_prob,
                                                               lasso_lambda=lambda_regul,
                                                               triplet_lambda=triplet_lambda)
    # optimization operator
    train_op = tf.train.AdamOptimizer(learn_rate).minimize(loss)

    """ MUSE optimization """
    total_batch = int(n_sample / batch_size)
    with tf.device('/gpu:0'):
        with tf.Session(config=config) as sess:

            """ initialization of autoencoder architecture for MUSE """
            print('MUSE initialization')
            # global parameter initialization
            sess.run(tf.global_variables_initializer(), feed_dict={keep_prob: keep_prob_val,
                                                                   triplet_lambda: 0,
                                                                   triplet_margin: 0})

            for epoch in range(n_epochs_init):
                # randomly permute samples
                random_idx = np.random.permutation(n_sample)
                data_train_x = data_x[random_idx, :]
                data_train_y = data_y[random_idx, :]

                for i in range(total_batch):
                    # input data batches
                    offset = (i * batch_size) % (n_sample)
                    batch_x_input = data_train_x[offset:(offset + batch_size), :]
                    batch_y_input = data_train_y[offset:(offset + batch_size), :]

                    # initialize parameters without self-supervised loss (triplet_lambda=0)
                    sess.run(train_op,
                             feed_dict={x: batch_x_input,
                                        y: batch_y_input,
                                        ref_label_x: np.zeros(batch_x_input.shape[0]),
                                        ref_label_y: np.zeros(batch_y_input.shape[0]),
                                        keep_prob: keep_prob_val,
                                        triplet_lambda: 0,
                                        triplet_margin: 0})

                # calculate and print loss terms for current epoch
                if epoch % print_epochs == 0:
                    L_total, L_reconstruction, L_weight = \
                        sess.run((loss, reconstruction_error, weight_penalty),
                                 feed_dict={x: data_train_x,
                                            y: data_train_y,
                                            ref_label_x: np.zeros(data_train_x.shape[0]),  # no use as triplet_lambda=0
                                            ref_label_y: np.zeros(data_train_y.shape[0]),  # no use as triplet_lambda=0
                                            keep_prob: keep_prob_val,
                                            triplet_lambda: 0,
                                            triplet_margin: 0})

                    print(
                        "epoch: %d, \t total loss: %03.5f,\t reconstruction loss: %03.5f,\t sparse penalty: %03.5f"
                        % (epoch, L_total, L_reconstruction, L_weight))

            # estimate the margin for the triplet loss
            latent, reconstruct_x, reconstruct_y = \
                sess.run((z, x_hat, y_hat),
                         feed_dict={x: data_x,
                                    y: data_y,
                                    ref_label_x: np.zeros(data_x.shape[0]),
                                    ref_label_y: np.zeros(data_y.shape[0]),
                                    keep_prob: keep_prob_val,
                                    triplet_lambda: 0,
                                    triplet_margin: 0})
            latent_pd_matrix = pdist(latent, 'euclidean')
            latent_pd_sort = np.sort(latent_pd_matrix)
            select_top_n = np.int(latent_pd_sort.size * 0.2)
            margin_estimate = np.median(latent_pd_sort[-select_top_n:]) - np.median(latent_pd_sort[:select_top_n])

            # refine MUSE parameters with reference labels and triplet losses
            for epoch in range(n_epochs_init):
                # randomly permute samples
                random_idx = np.random.permutation(n_sample)
                data_train_x = data_x[random_idx, :]
                data_train_y = data_y[random_idx, :]
                label_train_x = label_x[random_idx]
                label_train_y = label_y[random_idx]

                for i in range(total_batch):
                    # data batches
                    offset = (i * batch_size) % (n_sample)
                    batch_x_input = data_train_x[offset:(offset + batch_size), :]
                    batch_y_input = data_train_y[offset:(offset + batch_size), :]
                    label_x_input = label_train_x[offset:(offset + batch_size)]
                    label_y_input = label_train_y[offset:(offset + batch_size)]

                    # refine parameters
                    sess.run(train_op,
                             feed_dict={x: batch_x_input,
                                        y: batch_y_input,
                                        ref_label_x: label_x_input,
                                        ref_label_y: label_y_input,
                                        keep_prob: keep_prob_val,
                                        triplet_lambda: lambda_super,
                                        triplet_margin: margin_estimate})

                # calculate loss on all input data for current epoch
                if epoch % print_epochs == 0:
                    L_total, L_reconstruction, L_weight, L_trip_x, L_trip_y = \
                        sess.run((loss, reconstruction_error, weight_penalty, trip_loss_x, trip_loss_y),
                                 feed_dict={x: data_train_x,
                                            y: data_train_y,
                                            ref_label_x: label_train_x,
                                            ref_label_y: label_train_y,
                                            keep_prob: keep_prob_val,
                                            triplet_lambda: lambda_super,
                                            triplet_margin: margin_estimate})

                    print(
                        "epoch: %d, \t total loss: %03.5f,\t reconstruction loss: %03.5f,\t sparse penalty: %03.5f,\t x triplet: %03.5f,\t y triplet: %03.5f"
                        % (epoch, L_total, L_reconstruction, L_weight, L_trip_x, L_trip_y))

            # update cluster labels based modality-specific latents
            latent_x, latent_y = \
                sess.run((encode_x, encode_y),
                         feed_dict={x: data_x,
                                    y: data_y,
                                    ref_label_x: label_x,
                                    ref_label_y: label_y,
                                    keep_prob: keep_prob_val,
                                    triplet_lambda: lambda_super,
                                    triplet_margin: margin_estimate})

            # update cluster labels using PhenoGraph
            label_x_update, _, _ = phenograph.cluster(latent_x)
            label_y_update, _, _ = phenograph.cluster(latent_y)
            print('Finish initialization of MUSE')

            ''' Training of MUSE '''
            for epoch in range(n_epochs):
                # randomly permute samples
                random_idx = np.random.permutation(n_sample)
                data_train_x = data_x[random_idx, :]
                data_train_y = data_y[random_idx, :]
                label_train_x = label_x_update[random_idx]
                label_train_y = label_y_update[random_idx]

                # loop over all batches
                for i in range(total_batch):
                    # batch data
                    offset = (i * batch_size) % (n_sample)
                    batch_x_input = data_train_x[offset:(offset + batch_size), :]
                    batch_y_input = data_train_y[offset:(offset + batch_size), :]
                    batch_label_x_input = label_train_x[offset:(offset + batch_size)]
                    batch_label_y_input = label_train_y[offset:(offset + batch_size)]

                    sess.run(train_op,
                             feed_dict={x: batch_x_input,
                                        y: batch_y_input,
                                        ref_label_x: batch_label_x_input,
                                        ref_label_y: batch_label_y_input,
                                        keep_prob: keep_prob_val,
                                        triplet_lambda: lambda_super,
                                        triplet_margin: margin_estimate})

                # calculate and print losses on whole training dataset
                if epoch % print_epochs == 0:
                    L_total, L_reconstruction, L_weight, L_trip_x, L_trip_y = \
                        sess.run((loss, reconstruction_error, weight_penalty, trip_loss_x, trip_loss_y),
                                 feed_dict={x: data_train_x,
                                            y: data_train_y,
                                            ref_label_x: label_train_x,
                                            ref_label_y: label_train_y,
                                            keep_prob: keep_prob_val,
                                            triplet_lambda: lambda_super,
                                            triplet_margin: margin_estimate})
                    # print cost every epoch
                    print(
                        "epoch: %d, \t total loss: %03.5f,\t reconstruction loss: %03.5f,\t sparse penalty: %03.5f,\t x triplet loss: %03.5f,\t y triplet loss: %03.5f"
                        % (epoch, L_total, L_reconstruction, L_weight, L_trip_x, L_trip_y))

                # update cluster labels based on new modality-specific latent representations
                if epoch % cluster_update_epoch == 0:
                    latent_x, latent_y = \
                        sess.run((encode_x, encode_y),
                                 feed_dict={x: data_x,
                                            y: data_y,
                                            ref_label_x: label_x,
                                            ref_label_y: label_y,
                                            keep_prob: keep_prob_val,
                                            triplet_lambda: lambda_super,
                                            triplet_margin: margin_estimate})

                    # use PhenoGraph to obtain cluster label
                    label_x_update, _, _ = phenograph.cluster(latent_x)
                    label_y_update, _, _ = phenograph.cluster(latent_y)

            """ MUSE output """
            latent, reconstruct_x, reconstruct_y, latent_x, latent_y = \
                sess.run((z, x_hat, y_hat, encode_x, encode_y),
                         feed_dict={x: data_x,
                                    y: data_y,
                                    ref_label_x: label_x,  # no effects to representations
                                    ref_label_y: label_y,  # no effects to representations
                                    keep_prob: keep_prob_val,
                                    triplet_lambda: lambda_super,
                                    triplet_margin: margin_estimate})

    return latent, reconstruct_x, reconstruct_y, latent_x, latent_y



from setuptools import setup

setup(
    name='muse_sc',
    version='0.0.4',
    description='MUSE is a deep learning approach characterizing tissue composition through combined analysis of morphologies and transcriptional states for single-cell spatial transcriptomics data.',
    license='MIT',
    packages=['muse_sc'],
    author='Feng Bao @ AltschulerWu-Lab',
    author_email='feng.bao@ucsf.edu',
    keywords=['single-cell analysis', 'deep learning', 'spatial transcriptomics'],
    url='https://github.com/AltschulerWu-Lab/MUSE'
)
